/*******************************************************************************
 * Copyright (c) 2007 E.D.Willink and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     E.D.Willink - initial API and implementation
 *******************************************************************************/
package org.eclipse.gmt.umlx.eqvt.tests.parser;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import junit.framework.AssertionFailedError;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.common.util.DiagnosticChain;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EValidator;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.gmt.umlx.cst.ProblemLog;
import org.eclipse.gmt.umlx.cst.ProblemLog.Problem;
import org.eclipse.gmt.umlx.cst.environment.IFileEnvironment;
import org.eclipse.gmt.umlx.cst.environment.ProblemHandlerDiagnosticChain;
import org.eclipse.gmt.umlx.cst.unparser.AbstractUnparser;
import org.eclipse.gmt.umlx.eqvt.tests.AbstractTestCase;
import org.eclipse.gmt.umlx.eqvt.tests.TestFormattingHelper;
import org.eclipse.gmt.umlx.eqvt.tests.tools.EcoreEquivalenceHelper;
import org.eclipse.gmt.umlx.eqvt.tests.tools.EquivalenceHelper;
import org.eclipse.ocl.internal.l10n.OCLMessages;
import org.eclipse.ocl.lpg.AbstractParser;
import org.eclipse.ocl.lpg.AbstractProblemHandler;
import org.eclipse.ocl.lpg.ProblemHandler;
import org.eclipse.ocl.lpg.StringProblemHandler;
import org.eclipse.qvt.declarative.ecore.mappings.MappingConfigurationException;
import org.eclipse.qvt.declarative.registry.standalone.FileHandle;

public abstract class AbstractParseTestCase extends AbstractTestCase
{
	private final class TestProblemHandler extends StringProblemHandler
	{
		private TestProblemHandler() {
			super(null);
		}

		/**
		 * Implements the interface, invoking <code>addProblem</code> with a line comprising
		 * <code>processingPhase-problemSeverity in processingContext; lineNumber : problemMessage</code>.
		 */
		@Override
		public void handleProblem(Severity problemSeverity, Phase processingPhase,
				String problemMessage, String processingContext, int startOffset, int endOffset) {
			AbstractParser parser = getParser();
			String file = parser.getFileName();
			int lineNumber = parser.getTokenAtCharacter(startOffset).getLine();
			String message = OCLMessages.bind(
					 "{0}-{1} in {2}; {3}:{4} : {5}", //OCLMessages.ProblemMessage_ERROR_,
					new Object[] {
						processingPhase != null ? processingPhase : "?", //$NON-NLS-1$
						problemSeverity != null ? problemSeverity : "?", //$NON-NLS-1$
						processingContext != null ? processingContext : "?", //$NON-NLS-1$
						file, lineNumber,
						problemMessage != null ? problemMessage : "?"}); //$NON-NLS-1$
			addProblem(message);
		}
	}

	public static class ProblemComparator implements Comparator<Problem>
	{
		public static final Comparator<? super Problem> INSTANCE = new ProblemComparator();

		public int compare(Problem o1, Problem o2) {
			return o1.problemMessage.compareTo(o2.problemMessage);
		}
	}

	/**
	 * Asserts that two objects are equal. If they are not
	 * an AssertionFailedError is thrown.
	 */
	static public void assertEquals(ProblemLog expectedProblemLog, ProblemLog actualProblemLog) {
		try {
			int actualTotalProblems = actualProblemLog.getAllProblems();
			int expectedTotalProblems = expectedProblemLog.getAllProblems();
			assertEquals("Total Problem count", expectedTotalProblems, actualTotalProblems);
			Set<ProblemHandler.Phase> actualProcessingPhases = actualProblemLog.getProcessingPhases();
			Set<ProblemHandler.Phase> expectedProcessingPhases = expectedProblemLog.getProcessingPhases();
			if (!actualProcessingPhases.equals(expectedProcessingPhases)) {
				assertEquals("Processing Phases", expectedProcessingPhases, actualProcessingPhases);
			}
			for (ProblemHandler.Phase processingPhase : actualProcessingPhases) {
				Map<ProblemHandler.Severity, List<Problem>> actualPhaseProblems = actualProblemLog.getProblems(processingPhase);
				Map<ProblemHandler.Severity, List<Problem>> expectedPhaseProblems = expectedProblemLog.getProblems(processingPhase);
				Set<ProblemHandler.Severity> actualSeverities = actualPhaseProblems.keySet();
				Set<ProblemHandler.Severity> expectedSeverities = expectedPhaseProblems.keySet();
				if (!actualSeverities.equals(expectedSeverities)) {
					assertEquals("Processing Severities", expectedSeverities, actualSeverities);
				}
				for (ProblemHandler.Severity severity : actualSeverities) {
					List<Problem> actualProblems = actualPhaseProblems.get(severity);
					List<Problem> expectedProblems = expectedPhaseProblems.get(severity);
					Collections.sort(actualProblems, ProblemComparator.INSTANCE);
					Collections.sort(expectedProblems, ProblemComparator.INSTANCE);
					int actualProblemCount = actualProblems.size();
					int expectedProblemCount = expectedProblems.size();
					assertEquals("Problem count", expectedProblemCount, actualProblemCount);
					for (int i = 0; i < actualProblemCount; i++) {
						Problem actualProblem = actualProblems.get(i);
						Problem expectedProblem = expectedProblems.get(i);
						assertEquals("Problem message", expectedProblem.problemMessage, actualProblem.problemMessage);
						assertEquals("Problem processingContext", expectedProblem.processingContext, actualProblem.processingContext);
						if (expectedProblem.startOffset >= 0)
							assertEquals("Problem startOffset", expectedProblem.startOffset, actualProblem.startOffset);
						if (expectedProblem.endOffset >= 0)
							assertEquals("Problem endOffset", expectedProblem.endOffset, actualProblem.endOffset);
					}
				}
			}
		} catch (AssertionFailedError e) {
			StringBuffer s = new StringBuffer();
			s.append(e.getMessage());
			s.append("\nExpected");
			expectedProblemLog.toString(s, "  ");
			s.append("\nActual");
			actualProblemLog.toString(s, "  ");
			throw new AssertionFailedError(s.toString());
		}
	}
	
	protected Diagnostician createDiagnostician() {
		return new Diagnostician()
		{
			@Override
			protected boolean doValidateContents(EObject object, DiagnosticChain diagnostics, Map<Object, Object> context) {
				if (object.eClass().getEPackage() == org.eclipse.ocl.ecore.EcorePackage.eINSTANCE)
					return true;		// OCL's ValidatorVisitor does its own child traversal
				else
					return super.doValidateContents(object, diagnostics, context);
			}
		};
	}
	
	protected abstract IFileEnvironment createEnvironment(String fileName) throws IOException, CoreException;
	protected abstract AbstractUnparser createUnparser(Resource referenceResource);
	
	protected abstract String getTextExtension();
	protected abstract String getXMLExtension(boolean generateEMOF);
	
	/*
	 * Parse to Ecore/EMOF, check for no errors and check against reference.
	 */
	protected void parserTest(String fileStem, boolean generateEMOF) throws IOException, CoreException, MappingConfigurationException {
		parserTest(fileStem, null, generateEMOF);
	}
	protected void parserTest(String fileStem, ProblemLog expectedProblems, boolean generateEMOF) throws IOException, CoreException, MappingConfigurationException {
		IFileEnvironment environment = createEnvironment(fileStem + "." + getTextExtension());
		AbstractProblemHandler actualProblems;
		if (expectedProblems != null)
			actualProblems = new ProblemLog();
		else
			actualProblems = new TestProblemHandler();
		environment.setProblemHandler(actualProblems);
		//
		Collection<? extends EObject> asts = environment.parseToAST(null, null);
//		environment.validate(asts);
//		actualProblems.flush(null);
		//
		String generatedName = fileStem + "." + getXMLExtension(generateEMOF);
		URI generatedURI = getProjectFileURI(generatedName);
		Resource ecoreResource = environment.createASTResource(asts, generatedURI);
//		XMIResource resource = new XMIResourceImpl(generatedURI);
//		resourceSet.getResources().add(resource);
//		XMIUtils.assignIds(resource, "ast");
		if (generateEMOF) {
			Resource emofResource = getMappingMetaDataRegistry().getAdapter(ecoreResource, generatedURI);
			emofResource.save(null);
		}
		else
			ecoreResource.save(null);
		environment.validate(ecoreResource);
		actualProblems.flush(null);
		//
		if (expectedProblems != null) {
			expectedProblems.setParser(environment.getParser());
			assertEquals(expectedProblems, (ProblemLog)actualProblems);
		}
		else {
			int errorCount = ((StringProblemHandler) actualProblems).getProblemCount();
			if (errorCount > 0) {
				StringBuffer s = new StringBuffer();
				s.append(errorCount + " parser errors\n");
				s.append(((StringProblemHandler) actualProblems).getProblemString());
				s.setLength(s.length()-1);
				fail(s.toString());
			}
		}
		//
		String referenceName = fileStem + ".reference." + getXMLExtension(generateEMOF);
		URI projectURI = getProjectURI();
		URI referenceURI = getProjectFileURI(referenceName);
		EquivalenceHelper equivalenceHelper = /*generateEMOF ? EMOFEquivalenceHelper.INSTANCE :*/ EcoreEquivalenceHelper.INSTANCE;
		assertSameModel(equivalenceHelper, projectURI, ecoreResource, referenceURI, null);
	}

	protected void problemTest(String fileStem, ProblemLog expectedProblems) throws IOException, CoreException {
		ProblemLog actualProblems = new ProblemLog();
		IFileEnvironment environment = createEnvironment(fileStem + "." + getTextExtension());
		environment.setProblemHandler(actualProblems);
		Collection<? extends EObject> asts = environment.parseToAST(null, null);
		URI uri = URI.createFileURI(fileStem + "." + getTextExtension());
		Resource resource = environment.createASTResource(asts, uri);
		environment.validate(resource);
		actualProblems.flush(null);
		expectedProblems.setParser(environment.getParser());
		assertEquals(expectedProblems, actualProblems);
	}

	/*
	 * Load an Ecore/EMOF file, unparse to text, reparse to Ecore/EMOF, check for no errors and check against original.
	 */
	protected void unparserTest(String fileStem, boolean generateEMOF) throws IOException, CoreException, MappingConfigurationException {
		getMappingMetaDataRegistry().initialize();
		String referenceName = fileStem + ".reference." + getXMLExtension(generateEMOF);
		URI loadURI = getProjectFileURI(referenceName);
		Resource referenceResource = resourceSet.getResource(loadURI, true);;
		//
		AbstractUnparser unParser = createUnparser(referenceResource);
		unParser.unparse();
		String unparsedString = unParser.getString();
		//
		FileHandle fileHandle = getFileHandle(fileStem + ".unparsed." + getTextExtension());
		FileWriter fileWriter = new FileWriter(fileHandle.getFile());
		fileWriter.append(unparsedString);
		fileWriter.close();
		//
		IFileEnvironment environment = createEnvironment(fileStem + ".unparsed." + getTextExtension());
		StringProblemHandler problemHandler = new StringProblemHandler(null);
		environment.setProblemHandler(problemHandler);
		//
		Collection<? extends EObject> asts = environment.parseToAST(null, null);
//		environment.validate(asts);
//		problemHandler.flush(null);
		//
		int errorCount = problemHandler.getProblemCount();
		if (errorCount > 0)
			System.out.println(problemHandler.getProblemString());
		assertEquals("Parser errors", 0, errorCount);
		//
		String generatedName = fileStem + ".unparsed." + getXMLExtension(generateEMOF);
		URI generatedURI = getProjectFileURI(generatedName);
		Resource ecoreResource = environment.createASTResource(asts, generatedURI);
//		XMIUtils.assignIds(resource, "ast");
		if (generateEMOF) {
			Resource emofResource = getMappingMetaDataRegistry().getAdapter(ecoreResource, generatedURI);
			emofResource.save(null);
		}
		else
			ecoreResource.save(null);
		environment.validate(ecoreResource);
		problemHandler.flush(null);
		EquivalenceHelper equivalenceHelper = /* generateEMOF ? EMOFEquivalenceHelper.INSTANCE :*/ EcoreEquivalenceHelper.INSTANCE;
		assertSameModel(equivalenceHelper, getProjectURI(), ecoreResource, referenceResource, null);
	}

	protected void validationTest(String fileStem, ProblemLog expectedProblems, boolean generateEMOF) throws IOException, CoreException {
		AbstractProblemHandler actualProblems;
		if (expectedProblems != null)
			actualProblems = new ProblemLog();
		else
			actualProblems = new TestProblemHandler();
		String referenceName = fileStem + ".reference." + getXMLExtension(generateEMOF);
		URI loadURI = getProjectFileURI(referenceName);
		Resource resource = resourceSet.getResource(loadURI, true);
		DiagnosticChain diagnosticChain = new ProblemHandlerDiagnosticChain(actualProblems);
		Diagnostician diagnostician = createDiagnostician();
	    Map<Object, Object> context = new HashMap<Object, Object>();
	    context.put(EValidator.SubstitutionLabelProvider.class, TestFormattingHelper.INSTANCE);
	    context.put(EValidator.class, diagnostician);
		for (EObject eObject : resource.getContents()) {
			diagnostician.validate(eObject, diagnosticChain, context);
		}
		if (expectedProblems != null) {
//			expectedProblems.setParser(environment.getParser());
			assertEquals(expectedProblems, (ProblemLog)actualProblems);
		}
		else {
			int errorCount = ((StringProblemHandler) actualProblems).getProblemCount();
			if (errorCount > 0) {
				StringBuffer s = new StringBuffer();
				s.append(errorCount + " validation errors\n");
				s.append(((StringProblemHandler) actualProblems).getProblemString());
				s.setLength(s.length()-1);
				fail(s.toString());
			}
		}
	}
}
